#' process one EDF file
#'
#' Interpolate ECG signal to to reach at least \code{target.rate} samples per second (default: 200).
#' The upsampling factor \code{up} is chosen automatically as smallest integer satisfying that condition unless specified explicitely.
#'
#' @param file name of file, may include path
#' @param method indicate interpolation method (passed to \code{\link[signal]{interp1}})
#' @param columns number of columns in output file (1: interpolated signal only, 2: time points and interpolated signal).
#'                Automatically adjusted to nearest accepted value.
#' @param up integer factor by which to upsample, determined automatically if \code{NULL}.
#' @param target.rate minimum resulting sampling rate after interpolation
#' @param batch calling function may set batch to \code{TRUE} to indicate that
#'           errors should be reported as warnings so that the batch process
#'           can continue on other files.
#' @param verbose feedback?
#' @param ... passed on (interpolation, name conversion)
#'
#' @return list with elements
#' @export
#' @import edfReader
#' @import signal
#' @import utils
#'
#' @examples
#' # process.edf('test.edf')
process.edf <- function(file,
                        method      = "spline",
                        columns     = 2,
                        up          = NULL,
                        sample      = 'EKG',
                        target.rate = 200,
                        batch       = FALSE,
                        verbose     = getOption('verbose'),
                        ...){
    stop.fun <- list(stop, warning)[[batch+1]]
    dir.name <- dirname(file)
    if(dir.name != '.'){
        if(dir.exists(dir.name)){
            od <- setwd(dir.name)
            on.exit(setwd(od))
            file <- basename(file)
            if(!file.exists(file)){
                stop.fun(' no file "', file, '"')
                return(invisible(NULL))
            }
        } else {
            stop.fun('Directory "', dir.name, '" does not exist.')
            return(invisible(NULL))
        }
    }

    ## read and process EDF header
    hd <- edfReader::readEdfHeader(file)
    if(is.character(sample)){
        smplnum <- which(grepl(sample, hd$sHeaders$label))
    } else {
        max.idx <- length(hd$sHeaders$label)
        if (sample > 0 && sample <= max.idx){
            smplnum <- sample
        } else {
            stop.fun('bad sample index for "', file, '", should be less than ', max.idx+1)
            return(invisible(NULL))
        }
    }
    ## check for EKG existence - stop if suspicious
    if((lsn <- length(smplnum)) != 1){
        if(lsn >1){
            smplinfo <- paste0('\n\t',
                               sprintf("%2d: %s",
                                       smplnum,
                                       hd$sHeaders$label[smplnum]),
                               collapse = '\t')
        }
        stop.fun(sprintf('%s EKG signal%s in "%s"%s',
                         ifelse(lsn == 0,
                                'NO',
                                'multiple'),
                         ifelse(lsn == 0,
                                '',
                                's'),
                         file,
                         ifelse(lsn>0,
                                smplinfo,
                                '')))
        return(invisible(NULL))
    }

    ## upsampling factor
    sRate <- hd$sHeaders$sRate[smplnum]
    if (is.null(up)){        # need to calculate upsampling factor
        up <- ceiling(target.rate/sRate)
        if(verbose){
            cat(sprintf("\tupsample by %d%s",
                        up,
                        ifelse(batch, '', '\n')))
        }
    } else {
        if((reached.rate <- up * sRate) <= target.rate){
            warning(sprintf("interpolation for %s only reaches %.1f samples per second\n",
                            file,
                            reached.rate))
        }
    }
    sig <- edfReader::readEdfSignals(hd, smplnum)$signal

    ## prepare for interpolation
    len.sig <- length(sig)
    sr <- sRate                                # original sampling rate
    isr <- up * sr                             # interpolated sampling rate

    ## interpolate signal
    in.t <- (1:len.sig-1)/sr
    out.t <- (2:(len.sig*2)-2)/isr
    isig <- signal::interp1(in.t,              # original time points
                            sig,               # original signal
                            out.t,             # times at which to interpolate
                            method = method,   # interpolaton method
                            ...)

    ## prepare output
    columns <- min(2, max(1, columns))         # make sure value is valid
    output <- if(columns == 1){
        isig
    } else {
        cbind(out.t,
              isig)
    }
    utils::write.table(output,
                       file = base2target(file, ...),
                       row.names = FALSE,
                       col.names = FALSE)
    return(invisible(list(raw.signal = sig,
                          raw.times = in.t,
                          int.signal = isig,
                          int.times = out.t,
                          file = file)))
}

#' Process all EDF files in given folder
#'
#' @param path path to folder, default '.'
#' @param pattern (regular expression) pattern to match in candidate filenames (default: ending in 'EDF')
#' @param force set to TRUE to overwrite existing output files
#' @param ... passed on
#'
#' @return none
#' @export
#'
#' @examples
#' process.folder()
process.folder <- function(path    = '.',
                           pattern = 'EDF$',
                           force   = FALSE,
                           ...){
    if (path != '.'){
        od <- setwd(path)
        on.exit(setwd(od))
    }
    ## find all EDF files in the current directory
    candidates <- dir(pattern     = pattern,
                      ignore.case = TRUE)

    ## No files found? stop
    if(length(candidates) == 0){
        warning("No EDF files found in ",
                getwd())
        return(invisible(NULL))
    }

    ## process each file in turn
    for (file in candidates){
        target.name <- base2target(file, ...)
        cat(paste0(ifelse(file.exists(target.name) && force, 'Rep', 'P'),
                   'rocessing ',
                   file,
                   ' (',
                   which(candidates == file),
                   '/',
                   length(candidates),
                   ')'))
        if (!file.exists(target.name) || force){
            ignore <- process.edf(file,
                                  batch = TRUE,
                                  ...)
        } else {
            cat('\talready processed')
        }
        cat('\n')
    }
}

#' Derive output file name from input file
#'
#' only the actual name is considered and in it \code{old.extension} is replaced with \code{new.extension}.
#'
#' @section CAVEAT:
#' If \code{old.extension} is not found, \code{new.extension} is appended to the unmodified input.
#' @param filename filename to convert
#' @param old.extension extension to replace
#' @param new.extension replacement extension
#'
#' @return target name
#' @export
#'
#' @examples
#' base2target('dir/test.EDF')
#' base2target('dir/test.XDF')  # no match with old.extension
base2target <- function(filename,
                        old.extension = "\\.edf",
                        new.extension = '_i.txt'){
    ## substitute in two steps to make sure new.extension is added even if
    ## old.extension is not foiund in filename
    base <- sub(old.extension,
                '',
                basename(filename),
                ignore.case = TRUE)
    return(paste0(base, new.extension))
}

#' Check plot
#'
#' take the output from \code{\link{process.edf}} to interactively look at original and interpolated signal
#' @param signal output from \code{\link{process.edf}}
#' @param ...
#'
#' @return none
#' @export
#' @impoert manipulate
#'
#' @examples
#' \dontrun{
#' check.plot(process.edf('test.edf))
#' }
check.plot <- function(signal,
                       ...){
    sigplot <- function(signal, m,s,ms, width.s, width.ms, ...){
        offset <- (60*m + s + ms/1000) / dt
        width <- (width.s + width.ms/1000) / dt
        rng <-    offset  + 0:(  width - 1)
        rg2 <- 2*(offset) + 0:(2*width - 1)
        if(max(rng)>n) {
            corr <- max(rng) - n
            rng <- rng - corr
            rg2 <- rg2 - 2*corr
        }
        plot(signal[['raw.times']][rng], signal[['raw.signal']][rng], col = 'red',
             main = signal[['file']], las = 1, xlab = 'time [s]', ylab = 'ECG [mV]',
             ylim = yrng, pch ='+')

        axis(3, at = 0:(mins+1)*60, labels = 0:(mins+1))
        lines(signal[['int.times']][rg2], signal[['int.signal']][rg2], col = 'blue')
        grid()
        abline(h=0,             col = 'grey')
        abline(v=0:(mins+1)*60, col = 'grey')
    }
    yrng <- range(unlist(signal[c('raw.signel', 'int.signal')]))
    dt <- diff(signal[['raw.times']][1:2])
    n <- length(signal[['raw.times']])
    secs <- n * dt
    mins <- secs %/% 60
    secs <- secs %% 60
    manipulate::manipulate(sigplot(signal, min, sec, msec, width.sec, width.ms, ...),
               min       = slider(0,mins),
               sec       = slider(0,59),
               msec      = slider(0,999),
               width.sec = slider(1,250),
               width.ms  = slider(0,999))
}


##########################

#' HRolga: A package for computating the notorious bar statistic.
#'
#' The foo package provides three categories of important functions:
#' foo, bar and baz.
#'
#' @section HRolga functions:
#' \describe{
#' \item{process.edf}{process one EDF file}
#' \item{process.folder}{process all EDF files in one folder}
#' \item{check.plot}{interactive signal/interpolation exploration}
#' }
#'
#' @docType package
#' @name HRolga
NULL
